#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

int
ignoreerrors(Display *disp, XErrorEvent *eev)
{
	return 0;
}

int
getWindowName(Display *disp, Window w, char *buf, size_t buflen)
{
	XTextProperty prop;
	int rv;

	rv = XGetTextProperty(disp, w, &prop, XA_WM_NAME);
	if (!rv)
		snprintf(buf, buflen, "unknown");
	else
		snprintf(buf, buflen, "%s", prop.value);
	return 0;
}

int
main(int argc, char **argv)
{
	Display *disp;
	int scr;
	Window w, root, parent, *children;
	int rv, nchildren;
	char name[256];
	XEvent ev;
	XFocusChangeEvent *fev = (XFocusChangeEvent *)&ev;
	XCreateWindowEvent *cev = (XCreateWindowEvent *)&ev;
	XDestroyWindowEvent *dev = (XDestroyWindowEvent *)&ev;

	disp = XOpenDisplay("");
	scr = DefaultScreen(disp);
	root = RootWindow(disp, scr);

	XSetErrorHandler(ignoreerrors);

	XSelectInput(disp, root, FocusChangeMask | SubstructureNotifyMask);

	rv = XQueryTree(disp, root, &w, &parent, &children, &nchildren);
	for (int i = 0; i < nchildren; i++)
		XSelectInput(disp, children[i], FocusChangeMask);
	XFree(children);

	while (1) {
		XNextEvent(disp, &ev);
		switch (ev.type) {
		case FocusIn:
			getWindowName(disp, fev->window, name, sizeof(name));
			printf("entering %ld %s\n", fev->window, name);
			break;
		case FocusOut:
			getWindowName(disp, fev->window, name, sizeof(name));
			printf("leaving %ld %s\n", fev->window, name);
			break;
		case CreateNotify:
			XSelectInput(disp, cev->window, FocusChangeMask);
			getWindowName(disp, cev->window, name, sizeof(name));
			printf("created %ld %s\n", cev->window, name);
			break;
		case DestroyNotify:
			/* won't work, but we can try */
			getWindowName(disp, dev->window, name, sizeof(name));
			printf("destroyed %ld %s\n", dev->window, name);
			break;
		default:
			break;
		}
	}
	return 0;
}

